/*
    Sample C implementation for HPS pitch detection algorithm
    Copyright (C) 2011  Avelino Herrera Morales

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdint.h>
#include <unistd.h>
#include "fft.h"


#define  SAMPLE_RATE      44100
#define  LOG_BUFFER_SIZE  16
#define  BUFFER_SIZE      (1 << LOG_BUFFER_SIZE)
#define  HPS_STEP_MAX     3


float input[BUFFER_SIZE];
complex fft_buffer[BUFFER_SIZE];
complex fft_buffer_hps[BUFFER_SIZE];
complex fft_accumulator[BUFFER_SIZE];
float output[BUFFER_SIZE];
float window[BUFFER_SIZE];


void read_float_buffer(float *in) {
	int i;
	for (i = 0; i < BUFFER_SIZE; i++) {
		int16_t aux;
		read(0, &aux, sizeof(int16_t));
		in[i] = ((float) aux) / 32768.0;
	}
}


void fft_accumulator_write(void) {
	int i;
	for (i = 0; i < BUFFER_SIZE; i++) {
		int16_t v = (int16_t) rint((double) fft_accumulator[i].re);
		write(1, &v, sizeof(int16_t));
	}
}


void buffer_copy_float_to_complex(float *in, complex *out) {
	int i;
	for (i = 0; i < BUFFER_SIZE; i++) {
		out[i].re = in[i];
		out[i].im = 0;
	}
}


void buffer_copy_complex(complex *from, complex *to, int step) {
	int i, j;
	for (i = 0, j = 0; i < BUFFER_SIZE; i += step, j++) {
		to[j].re = from[i].re;
		to[j].im = from[i].im;
	}
	for ( ; j < BUFFER_SIZE; j++) {
		to[j].re = 0;
		to[j].im = 0;
	}
}


void fft_accumulator_init(complex *d) {
	int i;
	for (i = 0; i < BUFFER_SIZE; i++) {
		d[i].re = 1;
		d[i].im = 0;
	}
}


float get_hz_from_index(int i) {           // 0 < i < (BUFFER_SIZE / 2)
	return (((float) i) * SAMPLE_RATE) / BUFFER_SIZE;
}


void window_apply(float *w, float *data) {
	int i;
	for (i = 0; i < BUFFER_SIZE; i++, data++, w++)
		*data = (*data) * (*w);
}


int main(void) {
	int step, i, max_i;
	float max;
	read_float_buffer(input);
	fill_hanning_window(window, BUFFER_SIZE);
	window_apply(window, input);
	buffer_copy_float_to_complex(input, fft_buffer);
	fft(fft_buffer, LOG_BUFFER_SIZE);
	fft_accumulator_init(fft_accumulator);
	for (step = 1; step < HPS_STEP_MAX; step++) {
		buffer_copy_complex(fft_buffer, fft_buffer_hps, step);
		for (i = 0; i < BUFFER_SIZE; i++) {
			complex c;
			complex_mul(&c, &fft_accumulator[i], &fft_buffer_hps[i]);
			fft_accumulator[i] = c;
		}
	}
	// find max abs value at fft_accumulator
	max = 0;
	max_i = 0;
	for (i = 1; i < (BUFFER_SIZE / 2); i++) {
		float v = complex_abs(&fft_accumulator[i]);
		if (v > max) {
			max = v;
			max_i = i;
		}
	}
	printf("maximum found at %.04f Hz\n", get_hz_from_index(max_i));
	return 0;
}
